
// C includes
#include <unistd.h>
#include <sys/types.h>
#ifndef USING_MINGW
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif
#include <sys/time.h>
#include <fcntl.h>

// C++ includes
#include <algorithm>
using namespace std;

// MythTV includes
#include "mythdbcon.h"
#include "mythverbose.h"
#include "cetonchannel.h"
#include "videosource.h"
#include "channelutil.h"
#include "cetonstreamhandler.h"

#define DEBUG_PID_FILTERS

#define LOC     QString("CetonChan(%1): ").arg(GetDevice())
#define LOC_ERR QString("CetonChan(%1), Error: ").arg(GetDevice())

CetonChannel::CetonChannel(TVRec *parent, const QString &device)
    : DTVChannel(parent),
      _device_id(device),           _stream_handler(NULL),
      _lock(QMutex::Recursive),
      tune_lock(QMutex::Recursive), hw_lock(QMutex::Recursive)
{
}

CetonChannel::~CetonChannel(void)
{
    Close();
}

bool CetonChannel::Open(void)
{
    VERBOSE(VB_CHANNEL, LOC + "Opening Ceton channel");

    QMutexLocker locker(&hw_lock);

    if (IsOpen())
        return true;

    _stream_handler = CetonStreamHandler::Get(_device_id);


    tunerType = DTVTunerType::kTunerTypeATSC;
    _tuner_types.push_back(tunerType);

    if (!InitializeInputs())
    {
        Close();
        return false;
    }

    return _stream_handler->IsConnected();
}

void CetonChannel::Close(void)
{
    VERBOSE(VB_CHANNEL, LOC + "Closing Ceton channel");

    if (!IsOpen())
        return; // this caller didn't have it open in the first place..

    CetonStreamHandler::Return(_stream_handler);
}

bool CetonChannel::EnterPowerSavingMode(void)
{
    if (IsOpen())
        return _stream_handler->EnterPowerSavingMode();
    else
        return true;
}

bool CetonChannel::IsOpen(void) const
{
      return _stream_handler;
}

bool CetonChannel::Init(
    QString &inputname, QString &startchannel, bool setchan)
{
    if (setchan && !IsOpen())
        Open();

    return ChannelBase::Init(inputname, startchannel, setchan);
}

bool CetonChannel::SetChannelByString(const QString &channum)
{
    QString loc = LOC + QString("SetChannelByString(%1)").arg(channum);
    QString loc_err = loc + ", Error: ";
    VERBOSE(VB_CHANNEL, loc);

    if (!Open())
    {
        VERBOSE(VB_IMPORTANT, loc_err + "Channel object "
                "will not open, cannot change channels.");

        return false;
    }

    QString inputName;
    if (!CheckChannel(channum, inputName))
    {
        VERBOSE(VB_IMPORTANT, loc_err +
                "CheckChannel failed.\n\t\t\tPlease verify the channel "
                "in the 'mythtv-setup' Channel Editor.");

        return false;
    }

    // If CheckChannel filled in the inputName then we need to
    // change inputs and return, since the act of changing
    // inputs will change the channel as well.
    if (!inputName.isEmpty())
        return SelectInput(inputName, channum, false);

    ClearDTVInfo();

    InputMap::const_iterator it = m_inputs.find(m_currentInputID);
    if (it == m_inputs.end())
        return false;

    uint mplexid_restriction;
    if (!IsInputAvailable(m_currentInputID, mplexid_restriction))
        return false;

    if (Aborted())
        return false;

    // Fetch tuning data from the database.
    QString tvformat, modulation, freqtable, freqid, si_std;
    int finetune;
    uint64_t frequency;
    int mpeg_prog_num;
    uint atsc_major, atsc_minor, mplexid, tsid, netid;

    if (!ChannelUtil::GetChannelData(
        (*it)->sourceid, channum,
        tvformat, modulation, freqtable, freqid,
        finetune, frequency,
        si_std, mpeg_prog_num, atsc_major, atsc_minor, tsid, netid,
        mplexid, m_commfree))
    {
        return false;
    }

    if ((*it)->externalChanger.isEmpty())
    {
		if(_stream_handler->IsCableCardInstalled())
		{
			if(!_stream_handler->TuneCableCard(atsc_major, mpeg_prog_num))
			{
				VERBOSE(VB_IMPORTANT, LOC_ERR + "could not tune virtual channel");
				return false;
			}
		    SetSIStandard(si_std);
		}
		else
		{
			if(!_stream_handler->TuneQAM(frequency, modulation, mpeg_prog_num))
			{
				VERBOSE(VB_IMPORTANT, LOC_ERR + "could not tune QAM channel");
				return false;
			}
		}
    }
    else if (!ChangeExternalChannel(freqid))
        return false;

    // Set the current channum to the new channel's channum
    QString tmp = channum; tmp.detach();
    m_curchannelname = tmp;

    // Set the major and minor channel for any additional multiplex tuning
    SetDTVInfo(atsc_major, atsc_minor, netid, tsid, mpeg_prog_num);

    // Set this as the future start channel for this source
    QString tmpX = m_curchannelname; tmpX.detach();
    m_inputs[m_currentInputID]->startChanNum = tmpX;

    return true;
}

// documented in dtvchannel.h
bool CetonChannel::TuneMultiplex(uint mplexid, QString inputname)
{
    QString  modulation;
    QString  si_std;
    uint64_t frequency;
    uint     transportid;
    uint     dvb_networkid;

    if (!ChannelUtil::GetTuningParams(
            mplexid, modulation, frequency,
            transportid, dvb_networkid, si_std))
    {
        VERBOSE(VB_IMPORTANT, LOC_ERR + "TuneMultiplex(): " +
                QString("Could not find tuning parameters for multiplex %1.")
                .arg(mplexid));

        return false;
    }

    if (!Tune(frequency, inputname, modulation, 0, si_std))
        return false;

    return true;
}

bool CetonChannel::Tune(const DTVMultiplex &tuning, QString inputname)
{
    return Tune(tuning.frequency, inputname,
                tuning.modulation.toString(), 0, tuning.sistandard);
}

bool CetonChannel::Tune(uint frequency, QString /*input*/ input,
                       QString modulation, uint program, QString si_std)
{
	if(_stream_handler->TuneQAM(frequency, modulation, program))
	{
		SetSIStandard(si_std);
		return true;
	}

    return false;
}
