#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>

#include <QMutexLocker>
#include <QStringList>
#include <QUdpSocket>

#include "cetonrtp.h"
#include "mythverbose.h"

#define LOC QString("CetonRTP(%1-%2): ").arg(_ip).arg(_tuner)
#define LOC_WARNING QString("CetonRTP(%1-%2) Warning: ").arg(_ip).arg(_tuner)
#define LOC_ERR QString("CetonRTP(%1-%2) Error: ").arg(_ip).arg(_tuner)

#define RTPDATAGRAMSIZE 1328

CetonRTP::CetonRTP(const QString& ip, uint tuner, ushort listenPort,
                   ushort controlPort, QObject *parent):
    QObject(parent),
    _ip(ip),
    _tuner(tuner),
    _listenPort(listenPort),
    _rtpSocket(NULL),
    _acceptFrom(ip),
    _rtsp(ip, tuner, controlPort),
    _lastRead(0),
    _thisRead(0),
    _lastSequenceNumber(-1)
{
}

bool CetonRTP::initialize()
{
    QStringList options;
    if (!(_rtsp.doOptions(options)      && options.contains("OPTIONS")  &&
          options.contains("DESCRIBE")  && options.contains("SETUP")    &&
          options.contains("PLAY")      && options.contains("TEARDOWN")))
    {
        VERBOSE(VB_RECORD, LOC_ERR +
            "RTSP interface did not support the necessary options");
        return false;
    };

    if (!_rtpSocket.bind(_listenPort))
    {
        VERBOSE(VB_RECORD, LOC_ERR +
            QString("UDP socket failed to bind on port %1").arg(_listenPort));
        return false;
    };

    if (!_dummySocket.bind())
    {
        VERBOSE(VB_RECORD, LOC_ERR +
            QString("UDP dummy socket failed to bind on any port"));
        return false;
    };

    int socket = _rtpSocket.socketDescriptor();
    if (socket > 0)
    {
            int oldBufferSize;
            socklen_t bs_size = sizeof(oldBufferSize);
            getsockopt(socket,SOL_SOCKET, SO_RCVBUF, &oldBufferSize, &bs_size);

            int requestedBufferSize = 1024*1024*2;
            bs_size = sizeof(requestedBufferSize);
            if (setsockopt(socket,SOL_SOCKET, SO_RCVBUF, &requestedBufferSize,
                           sizeof(requestedBufferSize)) == -1)
                VERBOSE(VB_RECORD, LOC_WARNING +
                    QString("Unable to adjust socket receive buffer size"));

            int newBufferSize;
            bs_size = sizeof(newBufferSize);
            getsockopt(socket,SOL_SOCKET, SO_RCVBUF, &newBufferSize, &bs_size);
            VERBOSE(VB_RECORD, LOC +
                QString("Buffer size was %1 - requested %2 - now %3")
                .arg(oldBufferSize).arg(requestedBufferSize).arg(newBufferSize));
    };

    return true;
}

bool CetonRTP::beginStreaming()
{
    return _rtsp.doDescribe() &&
            _rtsp.doSetup(_rtpSocket.localPort(), _dummySocket.localPort()) &&
           _rtsp.doPlay();
};

bool CetonRTP::endStreaming()
{
    return _rtsp.doTeardown();
};


int CetonRTP::read(char *buffer, int bufferSize)
{
    processIncomingData();

    int count =_mpegBuffer.size();
    if (count > bufferSize) count = bufferSize;

    char* data = _mpegBuffer.data();
    memcpy(buffer, data, count);
    _mpegBuffer.remove(0, count);

    return count;
}

void CetonRTP::processIncomingData()
{
    QHostAddress receivedFrom;

    _lastRead = _thisRead;
    _thisRead = 0;

    int bufferSize = _readBuffer.size();

    while(_rtpSocket.hasPendingDatagrams())
    {
        int bytesWaiting = _rtpSocket.pendingDatagramSize();
        _readBuffer.resize(bufferSize+bytesWaiting);
        char* data = _readBuffer.data();

        int bytesRead = _rtpSocket.readDatagram(data+bufferSize,
                                                bytesWaiting, &receivedFrom);
        if (bytesRead != bytesWaiting)
            VERBOSE(VB_RECORD, LOC_ERR +
                QString("RTP packet error. Expected %1 bytes, received %2")
                .arg(bytesWaiting).arg(bytesRead));

        if (receivedFrom != _acceptFrom)
        {
            VERBOSE(VB_RECORD, LOC_WARNING +
                QString("Got packet from %1 instead of %2")
                .arg(receivedFrom.toString(), _acceptFrom.toString()));
            _readBuffer.resize(bufferSize);
            continue;
        };

        bufferSize += bytesWaiting;
        _thisRead += bytesWaiting;
     }

    char* data = _readBuffer.data();
    int offset = 0;
    while((offset+RTPDATAGRAMSIZE) <= bufferSize)
    {
        char* packet = &data[offset];
        char* payload;
        int payloadSize;
        int sequenceNumber;

        if (findRtpPayload(packet, payload, payloadSize, sequenceNumber))
        {
            _mpegBuffer.append(payload, payloadSize);

            if (_lastSequenceNumber != -1)
            {
                int expectedSequenceNumber = (_lastSequenceNumber+1) & 0x0FFFF;
                if (sequenceNumber != expectedSequenceNumber)
                    VERBOSE(VB_RECORD, LOC_WARNING +
                        QString("Sequence number went from %1 to %2 [last read=%3, this read=%4]")
                        .arg(_lastSequenceNumber).arg(sequenceNumber).arg(_lastRead).arg(_thisRead));
            };
            _lastSequenceNumber = sequenceNumber;
        }

        offset += RTPDATAGRAMSIZE;
    };

    _readBuffer.remove(0, offset);

}

bool CetonRTP::findRtpPayload(char* packet, char* &payload, int &payloadSize,
                              int &sequenceNumber)
{
    RtpFixedHeader* header = (RtpFixedHeader*)packet;

    if (header->version != 2)
    {
        VERBOSE(VB_RECORD, LOC_ERR +
            QString("Unhandled RTP version %1 packet (expected ver 2)")
            .arg(header->version));
        return false;
    }
    if (header->payloadtype != 33)
    {
        VERBOSE(VB_RECORD, LOC_ERR +
            QString("Error: can only process RTP packets with content type 33"));
        return false;
    }
    int headerSize = sizeof(RtpFixedHeader) + (4*header->csrccount);

    if (header->extension == 1)
    {
        ushort extHeaderLength = *(ushort*)(packet+headerSize+2);
        headerSize += 4 * (1 + extHeaderLength);
    }

    payload = packet + headerSize;
    payloadSize = 1328 - headerSize;
    sequenceNumber = ntohs(header->sequencenumber);

    if (header->padding)
    {
        int paddingSize = payload[1328];
        payloadSize -= paddingSize;
    }


    return true;
};
