#include <QStringList>
#include <QTcpSocket>

// MythTV includes
#include "cetonrtsp.h"
#include "mythverbose.h"


#define LOC QString("CetonRTSP(%1-%2): ").arg(_ip).arg(_tuner)
#define LOC_WARNING QString("CetonRTSP(%1-%2) Warning: ").arg(_ip).arg(_tuner)
#define LOC_ERR QString("CetonRTSP(%1-%2) Error: ").arg(_ip).arg(_tuner)

QMutex CetonRTSP::_rtspMutex;

CetonRTSP::CetonRTSP(const QString& ip, uint tuner, ushort port, QObject *parent) :
    QObject(parent),
    _ip(ip),
    _port(port),
    _tuner(tuner),
    _sequenceNumber(0),
    _sessionNumber(0)
{
    _requestUrl = QString("rtsp://%1:%2/cetonmpeg%3").arg(ip).arg(port).arg(tuner);
}

bool CetonRTSP::processRequest(const QString& method, const QStringList* headers)
{
    QMutexLocker locker(&_rtspMutex);
    QTcpSocket socket;
    socket.connectToHost(_ip, _port);

    QStringList requestHeaders;
    requestHeaders.append(QString("%1 %2 RTSP/1.0").arg(method, _requestUrl));
    requestHeaders.append(QString("User-Agent: MythTV Ceton Recorder"));
    requestHeaders.append(QString("CSeq: %1").arg(++_sequenceNumber));
    if (_sessionNumber > 0)
        requestHeaders.append(QString("Session: %1").arg(_sessionNumber));
    if (headers != NULL)
    {
        for(int i=0; i < headers->count(); i++)
        {
            QString header = headers->at(i);
            requestHeaders.append(header);
        }
    }
    requestHeaders.append(QString("\r\n"));
    QString request = requestHeaders.join("\r\n");


    //VERBOSE(VB_RECORD, LOC + QString("write: %1").arg(request));
    socket.write(request.toAscii());

    _responseHeaders.clear();
    _responseContent.clear();

    QRegExp firstLineRegex("^RTSP/1.0 (\\d+) ([^\r\n]+)", Qt::CaseSensitive, QRegExp::RegExp2);
    QRegExp headerRegex("^([^:]+):\\s*([^\\r\\n]+)", Qt::CaseSensitive, QRegExp::RegExp2);
    QRegExp blankLineRegex("^[\\r\\n]*$", Qt::CaseSensitive, QRegExp::RegExp2);

    bool firstLine = true;
    while(1)
    {
        if (!socket.canReadLine())
        {
            bool ready = socket.waitForReadyRead();
            if (!ready)
            {
                VERBOSE(VB_RECORD, LOC_ERR + "RTSP server did not respond");
                return false;
            }
            continue;
        };

        QString line = socket.readLine();
        //VERBOSE(VB_RECORD, LOC + QString("read: %1").arg(line));

        if (firstLine)
        {
            if (firstLineRegex.indexIn(line) == -1)
            {
                _responseCode = -1;
                _responseMessage = QString("Could not parse first line of response: %1").arg(line);
                return false;
            };

            QStringList parts = firstLineRegex.capturedTexts();
            _responseCode = parts.at(1).toInt();
            _responseMessage = parts.at(2);

            firstLine = false;
            continue;
        }

        if (blankLineRegex.indexIn(line) != -1) break;

        if (headerRegex.indexIn(line) == -1)
        {
            _responseCode = -1;
            _responseMessage = QString("Could not parse response header: %1").arg(line);
            return false;
        };
        QStringList parts = headerRegex.capturedTexts();
        _responseHeaders.insert(parts.at(1), parts.at(2));
    };

    QString cSeq = _responseHeaders.value("CSeq");
    if (cSeq != QString("%1").arg(_sequenceNumber))
        VERBOSE(VB_RECORD, LOC_WARNING + QString("Expected CSeq of %1 but got %2").arg(_sequenceNumber).arg(cSeq));

    _responseContent.clear();
    int contentLength = _responseHeaders.value("Content-Length").toInt();
    if (contentLength > 0)
    {
        _responseContent.resize(contentLength);
        char* data = _responseContent.data();
        int bytesRead = 0;
        while(bytesRead < contentLength)
        {
            if (socket.bytesAvailable() == 0)
                socket.waitForReadyRead();

            int count = socket.read(data+bytesRead, contentLength-bytesRead);
            if (count == -1)
            {
                _responseCode = -1;
                _responseMessage = QString("Could not read response content");
                return false;
            };
            bytesRead += count;
        }
    }
    return true;
};


bool CetonRTSP::doOptions(QStringList& options)
{
    if (!processRequest("OPTIONS")) return false;

    options = _responseHeaders.value("Public").split(",");
    return true;
};

bool CetonRTSP::doDescribe()
{
    if (!processRequest("DESCRIBE")) return false;

    if (!_responseContent.contains("m=video 0 RTP/AVP 33"))
    {
        VERBOSE(VB_RECORD, LOC_ERR + "expected content to be type \"m=video 0 RTP/AVP 33\" but it appears not to be");
        return false;
    };

    return true;
};

bool CetonRTSP::doSetup(ushort clientPort1, ushort clientPort2)
{
    QStringList extraHeaders;
    extraHeaders.append(QString("Transport: RTP/AVP;unicast;client_port=%1-%2").arg(clientPort1).arg(clientPort2));

    if (!processRequest("SETUP", &extraHeaders)) return false;

    _sessionNumber = _responseHeaders.value("Session").toInt();
    if (_sessionNumber == 0)
    {
        VERBOSE(VB_RECORD, LOC_ERR + "session number not found in SETUP response");
        return false;
    }

    return true;
};

bool CetonRTSP::doPlay()
{
    return processRequest("PLAY");
};

bool CetonRTSP::doTeardown()
{
    bool result = processRequest("TEARDOWN");
    _sessionNumber = 0;
    return result;
};
