
// POSIX headers
#include <pthread.h>
#include <fcntl.h>
#include <unistd.h>
#ifndef USING_MINGW
#include <sys/select.h>
#include <sys/ioctl.h>
#endif

// Qt headers
#include <QApplication>
#include <QCoreApplication>
#include <QString>
#include <QFile>
#include <QUrl>
#include <QBuffer>
#include <QHttp>

// MythTV headers
#include "cetonstreamhandler.h"
#include "cetonchannel.h"
#include "cetonrtp.h"
#include "dtvsignalmonitor.h"
#include "streamlisteners.h"
#include "mpegstreamdata.h"
#include "cardutil.h"

#include <netinet/in.h>

#define LOC      QString("CetonSH(%1): ").arg(_devicename)
#define LOC_WARN QString("CetonSH(%1) Warning: ").arg(_devicename)
#define LOC_ERR  QString("CetonSH(%1) Error: ").arg(_devicename)

QMap<QString,CetonStreamHandler*> CetonStreamHandler::_handlers;
QMap<QString,uint>               CetonStreamHandler::_handlers_refcnt;
QMutex                           CetonStreamHandler::_handlers_lock;

QMap<QString, bool> CetonStreamHandler::_info_queried;

CetonHttpRequest::CetonHttpRequest(const QString &host, const QString &method, const QString &script, const QUrl &params) :
    _error(false),
    _status_code(0),
    _response(""),
    _host(host),
    _method(method),
    _script(script),
    _params(params)
{

}

bool CetonHttpRequest::request()
{
    start();
    wait();
    return !_error;
}

void CetonHttpRequest::run()
 {
        QHttp http;
        http.setHost(_host);

        QByteArray request_params(_params.encodedQuery());

        if (_method == "GET")
        {
                QString path = _script + "?" + QString(request_params);
                QHttpRequestHeader header(_method, path);
                header.setValue("Host", _host);
                http.request(header);
        }
        else
        {
                QHttpRequestHeader header(_method, _script);
                header.setValue("Host", _host);
                header.setContentType("application/x-www-form-urlencoded");
                http.request(header, request_params);
        }


        while(http.hasPendingRequests() || http.currentId())
        {
                usleep(5000);
                qApp->processEvents();
        }

        if(http.error() != QHttp::NoError)
        {
                _error = true;
                _status_code = 0;
                _response = http.errorString();
                return;
        }

        QHttpResponseHeader resp_header = http.lastResponse();
        if(!resp_header.isValid())
        {
                _error = true;
                _status_code = 0;
                _response = "Completed but response object was not valid";
                return;
        }

        _error = false;
        _status_code = resp_header.statusCode();
        _response = QString(http.readAll());

        return;
}



CetonStreamHandler *CetonStreamHandler::Get(const QString &devname)
{
    QMutexLocker locker(&_handlers_lock);

    QString devkey = devname.toUpper();

    QMap<QString,CetonStreamHandler*>::iterator it = _handlers.find(devkey);

    if (it == _handlers.end())
    {
        CetonStreamHandler *newhandler = new CetonStreamHandler(devkey);
        newhandler->Open();
        _handlers[devkey] = newhandler;
        _handlers_refcnt[devkey] = 1;

        VERBOSE(VB_RECORD,
                QString("CetonSH: Creating new stream handler %1 for %2")
                .arg(devkey).arg(devname));
    }
    else
    {
        _handlers_refcnt[devkey]++;
        uint rcount = _handlers_refcnt[devkey];
        VERBOSE(VB_RECORD,
                QString("CetonSH: Using existing stream handler %1 for %2")
                .arg(devkey)
                .arg(devname) + QString(" (%1 in use)").arg(rcount));
    }

    return _handlers[devkey];
}

void CetonStreamHandler::Return(CetonStreamHandler * & ref)
{
    QMutexLocker locker(&_handlers_lock);

    QString devname = ref->_devicename;

    QMap<QString,uint>::iterator rit = _handlers_refcnt.find(devname);
    if (rit == _handlers_refcnt.end())
        return;

    if (*rit > 1)
    {
        ref = NULL;
        (*rit)--;
        return;
    }

    QMap<QString,CetonStreamHandler*>::iterator it = _handlers.find(devname);
    if ((it != _handlers.end()) && (*it == ref))
    {
        VERBOSE(VB_RECORD, QString("CetonSH: Closing handler for %1")
                           .arg(devname));
        ref->Close();
        delete *it;
        _handlers.erase(it);
    }
    else
    {
        VERBOSE(VB_IMPORTANT,
                QString("CetonSH Error: Couldn't find handler for %1")
                .arg(devname));
    }

    _handlers_refcnt.erase(rit);
    ref = NULL;
}

CetonStreamHandler::CetonStreamHandler(const QString &devicename) :
    _devicename(devicename),
    _connected(false),
    _valid(false),
    _start_stop_lock(QMutex::Recursive),
    _running(false),
    _reader_thread(),
    _listener_lock(QMutex::Recursive)
{
    VERBOSE(VB_RECORD, LOC + QString("CetonStreamHandler(%1)").arg(devicename));

    QStringList parts = devicename.split("-");
    if (parts.size() != 2)
    {
        VERBOSE(VB_IMPORTANT, LOC_ERR + QString("Invalid device id %1").arg(devicename));
    	return;
    }
    _ip_address = parts.at(0);

    QStringList tuner_parts = parts.at(1).split(".");
    if (tuner_parts.size() == 1)
    {
        _using_rtp = false;
        _card = 0;
        _tuner = tuner_parts.at(0).toUInt() - 1;
        VERBOSE(VB_RECORD, LOC + QString("Using old style CETON definition from original patch"));
    }
    else if (tuner_parts.size() == 2)
    {
        _using_rtp = (tuner_parts.at(0) == "RTP");
        _card = tuner_parts.at(0).toUInt();
        _tuner = tuner_parts.at(1).toUInt();
    }
    else
    {
        VERBOSE(VB_RECORD, LOC_ERR +
            QString("Invalid device id %1").arg(_devicename));
        return;
    }

    if (GetVar("diag", "Host_IP_Address") == "")
    {
        VERBOSE(VB_RECORD, LOC_ERR +
            QString("Ceton tuner does not seem to be available at IP"));
            return;
    }

    if (!_using_rtp)
    {
        _device_path = QString("/dev/ceton/ctn91xx_mpeg%1_%2")
            .arg(_card).arg(_tuner);

        if (!QFile(_device_path).exists())
        {
            VERBOSE(VB_IMPORTANT, LOC_ERR + QString("Tuner device unavailable"));
            return;
        }
    }

    _valid = true;

    QString cardstatus = GetVar("cas", "CardStatus");
    _using_cablecard = cardstatus == "Inserted";

    if (!_info_queried.contains(_ip_address))
    {
        QString sernum = GetVar("diag", "Host_Serial_Number");
        QString firmware_ver = GetVar("diag", "Host_Firmware");
        QString hardware_ver = GetVar("diag", "Hardware_Revision");

        VERBOSE(VB_RECORD, LOC + QString("Ceton device initialized. SN: %1, Firmware ver. %2, Hardware ver. %3").arg(sernum).arg(firmware_ver).arg(hardware_ver));

        if (_using_cablecard)
        {
            QString brand = GetVar("cas", "CardManufacturer");
            QString auth = GetVar("cas", "CardAuthorization");

            VERBOSE(VB_RECORD, LOC + QString("Cable card installed (%1) - %2").arg(brand).arg(auth));
        }
        else
        {
            VERBOSE(VB_RECORD, LOC + QString("Cable card NOT installed (operating in QAM tuner mode)"));
        }

        _info_queried.insert(_ip_address, true);
    };
}

CetonStreamHandler::~CetonStreamHandler()
{
    if (!_stream_data_list.empty())
    {
        VERBOSE(VB_IMPORTANT, LOC_ERR + "dtor & _stream_data_list not empty");
    }
}

void CetonStreamHandler::AddListener(MPEGStreamData *data)
{
    VERBOSE(VB_RECORD, LOC + "AddListener("<<data<<") -- begin");
    if (!data)
    {
        VERBOSE(VB_IMPORTANT, LOC_ERR +
                "AddListener("<<data<<") -- null data");
        return;
    }

    _listener_lock.lock();

    VERBOSE(VB_RECORD, LOC + "AddListener("<<data<<") -- locked");

    _stream_data_list.push_back(data);

    _listener_lock.unlock();

    Start();

    VERBOSE(VB_RECORD, LOC + "AddListener("<<data<<") -- end");
}

void CetonStreamHandler::RemoveListener(MPEGStreamData *data)
{
    VERBOSE(VB_RECORD, LOC + "RemoveListener("<<data<<") -- begin");
    if (!data)
    {
        VERBOSE(VB_IMPORTANT, LOC_ERR +
                "RemoveListener("<<data<<") -- null data");
        return;
    }

    _listener_lock.lock();

    VERBOSE(VB_RECORD, LOC + "RemoveListener("<<data<<") -- locked");

    vector<MPEGStreamData*>::iterator it =
        find(_stream_data_list.begin(), _stream_data_list.end(), data);

    if (it != _stream_data_list.end())
        _stream_data_list.erase(it);

    if (_stream_data_list.empty())
    {
        _listener_lock.unlock();
        Stop();
    }
    else
    {
        _listener_lock.unlock();
    }

    VERBOSE(VB_RECORD, LOC + "RemoveListener("<<data<<") -- end");
}

void *run_ceton_stream_handler_thunk(void *param)
{
    CetonStreamHandler *mon = (CetonStreamHandler*) param;
    mon->Run();
    return NULL;
}

void CetonStreamHandler::Start(void)
{
    QMutexLocker locker(&_start_stop_lock);

    _eit_pids.clear();

    if (!IsRunning())
    {
        QMutex is_running_lock;
        int rval = pthread_create(&_reader_thread, NULL,
                                  run_ceton_stream_handler_thunk, this);

        if (0 != rval)
        {
            VERBOSE(VB_IMPORTANT, LOC_ERR +
                    "Start: Failed to create thread." + ENO);
            return;
        }

        is_running_lock.lock();
        while (!IsRunning())
        {
            _running_state_changed.wait(&is_running_lock, 100);
        }
    }
}

void CetonStreamHandler::Stop(void)
{
    QMutexLocker locker(&_start_stop_lock);

    if (IsRunning())
    {
        SetRunning(false);
        pthread_join(_reader_thread, NULL);
    }
}

void CetonStreamHandler::Run(void)
{
	VERBOSE(VB_CHANNEL, LOC + QString("Run()"));
    SetRunning(true);
    RunTS();
}

void CetonStreamHandler::RunTS(void)
{
    QFile file(_device_path);
    CetonRTP rtp(_ip_address, _tuner);

    if (_using_rtp)
    {
        if (!(rtp.initialize() && rtp.beginStreaming()))
        {
            VERBOSE(VB_RECORD, LOC_ERR +
                "Starting recording (RTP initialization failed). Aborting.");
            return;
        };
    }
    else
    {
        if (!file.open(QIODevice::ReadOnly))
        {
            VERBOSE(VB_RECORD, LOC_ERR +
                "Starting recording (file open failed). Aborting.");
            return;
        };

        int flags = fcntl(file.handle(), F_GETFL, 0);
        if (flags == -1) flags = 0;
        fcntl(file.handle(), F_SETFL, flags | O_NONBLOCK);
    }

    VERBOSE(VB_RECORD, LOC + "RunTS(): begin");

    _read_timer.start();

    #define VIDEO_DATA_PACKET_SIZE 188
    int buffer_size = (64 * 1024); // read about 64KB
    buffer_size /= VIDEO_DATA_PACKET_SIZE;
    buffer_size *= VIDEO_DATA_PACKET_SIZE;
    char buffer[buffer_size];

    bool _error = false;
    int remainder = 0;
    while (IsRunning() && !_error)
    {
        int bytes_read;
        if (_using_rtp)
            bytes_read = rtp.read(buffer, buffer_size);
        else
            bytes_read = file.read(buffer, buffer_size);

        if (bytes_read <= 0)
        {
            if (_read_timer.elapsed() >= 5000)
            {
                VERBOSE(VB_RECORD, LOC +
                    QString("No data received for 5 seconds...checking tuning"));
                if (!VerifyTuning())
                    RepeatTuning();
                _read_timer.start();
            }

            usleep(5000);
            continue;
        }

        _read_timer.start();


        _listener_lock.lock();

        if (_stream_data_list.empty())
        {
            _listener_lock.unlock();
            VERBOSE(VB_RECORD, LOC + "RunTS(): _stream_data_list is empty....skipping");
            continue;
        }

        for (uint i = 0; i < _stream_data_list.size(); i++)
        {
            remainder = _stream_data_list[i]->ProcessData(
            		(unsigned char*)buffer, bytes_read);
        }

        _listener_lock.unlock();
        if (remainder != 0)
        {
            VERBOSE(VB_RECORD, LOC +
                    QString("RunTS(): data_length = %1 remainder = %2")
                    .arg(bytes_read).arg(remainder));
        }
    }
    VERBOSE(VB_RECORD, LOC + "RunTS(): " + "shutdown");

    if (_using_rtp)
        rtp.endStreaming();
    else
        file.close();

    VERBOSE(VB_RECORD, LOC + "RunTS(): " + "end");

    SetRunning(false);
}

void CetonStreamHandler::SetRunning(bool is_running)
{
    _running = is_running;
    _running_state_changed.wakeAll();
}

bool CetonStreamHandler::Open(void)
{
    return Connect();
}

void CetonStreamHandler::Close(void)
{
    if (_connected)
    {
    	TunerOff();
        _connected = false;
    }
}

bool CetonStreamHandler::Connect(void)
{
	if (!_valid)
			return false;

	_connected = true;
    return true;
}

bool CetonStreamHandler::EnterPowerSavingMode(void)
{
    QMutexLocker locker(&_listener_lock);

    if (!_stream_data_list.empty())
    {
        VERBOSE(VB_RECORD, LOC + "Ignoring request - video streaming active");
        return false;
    }
    else
    {
        locker.unlock(); // _listener_lock
        TunerOff();
        return true;
    }
}

bool CetonStreamHandler::IsConnected(void) const
{
	return _connected;
}

bool CetonStreamHandler::TunerOff()
{
    bool result = TuneFrequency(0, "qam_256");
    if (result && _using_cablecard)
            result = TuneVChannel(0);

    return result;
}

bool CetonStreamHandler::TuneQAM(uint frequency, const QString &modulation, uint program)
{
    return TuneFrequency(frequency, modulation) && TuneProgram(program);
}

bool CetonStreamHandler::TuneCableCard(uint vchannel, int &program)
{
    bool result = TuneVChannel(vchannel);
    if (result)
        program = GetProgramNumber();

    return result;
}

bool CetonStreamHandler::TuneFrequency(uint frequency, const QString &modulation)
{
    VERBOSE(VB_RECORD, LOC + QString("TuneFrequency(%1, %2)")
        .arg(frequency).arg(modulation));

    if (frequency >= 100000000)
        frequency /= 1000;

    QString modulation_id = (modulation == "qam_256") ? "2" :
                            (modulation == "qam_64")  ? "0" :
                            (modulation == "ntsc-m") ? "4" :
                            (modulation == "8vsb")   ? "6" :
                                                       "";
    if (modulation_id == "")
        return false;

    _last_frequency = frequency;
    _last_modulation = modulation;

    QUrl params;
    params.addQueryItem("instance_id", QString::number(_tuner));
    params.addQueryItem("frequency", QString::number(frequency));
    params.addQueryItem("modulation",modulation_id);
    params.addQueryItem("tuner","1");
    params.addQueryItem("demod","1");
    params.addQueryItem("rst_chnl","0");
    params.addQueryItem("force_tune","0");

    QString response;
    uint status;
    bool result =  HttpRequest("POST", "/tune_request.cgi", params, response, status);
    if (!result)
    {
        VERBOSE(VB_RECORD, LOC_ERR +
            QString("TuneFrequency() - HTTP status = %1 - response = %2").arg(status).arg(response));
    };
    return result;
}

bool CetonStreamHandler::TuneProgram(uint program)
{
    VERBOSE(VB_RECORD, LOC + QString("TuneProgram(%1)").arg(program));

    QStringList program_list = GetProgramList();
    if (!program_list.contains(QString::number(program)))
    {
        VERBOSE(VB_IMPORTANT, LOC_ERR + 
            QString("TuneProgram(%1) - Requested program not in the program list").arg(program));
        return false;
    };

    _last_program = program;

    QUrl params;
    params.addQueryItem("instance_id", QString::number(_tuner));
    params.addQueryItem("program", QString::number(program));

    QString response;
    uint status;
    bool result =  HttpRequest("POST", "/program_request.cgi", params, response, status);
    if (!result)
    {
        VERBOSE(VB_RECORD, LOC_ERR +
            QString("TuneProgram() - HTTP status = %1 - response = %2").arg(status).arg(response));
    };
    return result;
}

bool CetonStreamHandler::TuneVChannel(uint vchannel)
{
    VERBOSE(VB_RECORD, LOC + QString("TuneVChannel(%1)").arg(vchannel));

    _last_vchannel = vchannel;

    QUrl params;
    params.addQueryItem("instance_id", QString::number(_tuner));
    params.addQueryItem("channel", QString::number(vchannel));

    QString response;
    uint status;
    bool result =  HttpRequest("POST", "/channel_request.cgi", params, response, status);
    if (!result)
    {
        VERBOSE(VB_RECORD, LOC_ERR +
            QString("TuneVChannel() - HTTP status = %1 - response = %2").arg(status).arg(response));
    };
    return result;
}

uint CetonStreamHandler::GetProgramNumber()
{
    for(int i = 1; i <= 5; i++)
    {
        QString prog = GetVar("mux", "ProgramNumber");
        VERBOSE(VB_RECORD, LOC + QString("GetProgramNumber() got %1 on attempt %2").arg(prog).arg(i));

        uint prognum = prog.toUInt();
        if (prognum != 0) 
            return prognum;

        usleep(100000);
    };

    VERBOSE(VB_IMPORTANT, LOC_ERR + QString("Error: GetProgramNumber() failed to get a non-zero program number"));
    return 0;
}

bool CetonStreamHandler::VerifyTuning()
{
    if (IsCableCardInstalled())
    {
        uint prog = GetVar("mux", "ProgramNumber").toUInt();
        if (prog == 0)
        {
            VERBOSE(VB_RECORD, LOC + "VerifyTuning detected program = 0");
            return false;
        }
    }
    else
    {
        uint frequency = GetVar("tuner","Frequency").toUInt();
        if (frequency != _last_frequency)
        {
            VERBOSE(VB_RECORD, LOC + "VerifyTuning detected wrong frequency");
            return false;
        }

        QString modulation = GetVar("tuner","Modulation");
        if (modulation.toUpper() != _last_modulation.toUpper())
        {
            VERBOSE(VB_RECORD, LOC + "VerifyTuning detected wrong modulation");
            return false;
        }

        uint program = GetVar("mux","ProgramNumber").toUInt();
        if (program != _last_program)
        {
            VERBOSE(VB_RECORD, LOC + "VerifyTuning detected wrong program");
            return false;
        }
    }

    QString carrier_lock = GetVar("tuner","CarrierLock");
    if (carrier_lock != "1")
    {
        VERBOSE(VB_RECORD, LOC + "VerifyTuning detected no carrier lock");
        return false;
    }

    QString pcr_lock = GetVar("tuner","PCRLock");
    if (pcr_lock != "1")
    {
        VERBOSE(VB_RECORD, LOC + "VerifyTuning detected no PCR lock");
        return false;
    }

    return true;
}

void CetonStreamHandler::RepeatTuning()
{
    if (IsCableCardInstalled())
    {
        TuneVChannel(_last_vchannel);
    }
    else
    {
        TuneFrequency(_last_frequency, _last_modulation);
        TuneProgram(_last_program);
    }
}

QString CetonStreamHandler::GetVar(const QString &section, const QString &variable)
{
    QString LOC_FUNC = QString("GetVar(%1,%2,%3,%4) - ").arg(_ip_address).arg(_tuner).arg(section,variable);

    QUrl params;
    params.addQueryItem("i", QString::number(_tuner));
    params.addQueryItem("s", section);
    params.addQueryItem("v", variable);

    QString response;
    uint status;
    if (!HttpRequest("GET", "/get_var.json", params, response, status))
    {
        VERBOSE(VB_RECORD, LOC + LOC_FUNC + QString("HttpRequest failed - %1").arg(response));
        return QString();
    }

    QRegExp regex("^\\{ \"?result\"?: \"(.*)\" \\}$");
    if (regex.indexIn(response) == -1)
    {
        VERBOSE(VB_RECORD, LOC + LOC_FUNC + QString("unexpected http response: -->%1<--").arg(response));
        return QString();
    }

    QString result = regex.cap(1);
    VERBOSE(VB_RECORD, LOC + LOC_FUNC + QString("got: -->%1<--").arg(result));
    return result;
}

QStringList CetonStreamHandler::GetProgramList()
{
    QString LOC_FUNC = QString("GetProgramList(%1,%2) - ").arg(_ip_address).arg(_tuner);

    QUrl params;
    params.addQueryItem("i", QString::number(_tuner));

    QString response;
    uint status;
    if (!HttpRequest("GET", "/get_pat.json", params, response, status))
    {
        VERBOSE(VB_RECORD, LOC + LOC_FUNC + QString("HttpRequest failed - %1").arg(response));
        return QStringList();
    }

    QRegExp regex("^\\{ \"?length\"?: \\d+(, \"?results\"?: \\[ (.*) \\])? \\}$");
    if (regex.indexIn(response) == -1)
    {
        VERBOSE(VB_RECORD, LOC + LOC_FUNC + QString("returned unexpected output: -->%1<--").arg(response));
        return QStringList();
    }

    VERBOSE(VB_RECORD, LOC + LOC_FUNC + QString("got: -->%1<--").arg(regex.cap(2)));
    return regex.cap(2).split(", ");
};

bool CetonStreamHandler::HttpRequest(const QString& method, const QString &script, const QUrl &params,
                QString &response, uint &status_code)
{
    CetonHttpRequest http(_ip_address, method, script, params);
    bool result = http.request();

    int status;
    response = http.getResult(status);

    if(!result)
    {
        VERBOSE(VB_IMPORTANT, LOC_ERR + QString("HTTP %1 request failed: %2 - %3").arg(method).arg(status).arg(response));
        return false;
    }

    return true;
}
